# Various utilities related to date handling

from datetime import date as Date
from datetime import datetime as DateTime
from datetime import timedelta as TimeDelta
import re
import functools

today = Date.today

FORMAT = re.compile(r'^(\d{4})(?:\-|\/)(\d{1,2})(?:\-|\/)(\d{1,2})$')

def absolute_date(info):
    if isinstance(info, Date):
        return info
    elif isinstance(info, tuple):
        return Date(*info)
    elif isinstance(info, str) and FORMAT.match(info):
        return Date(*(int(x) for x in FORMAT.match(info).groups()))
    raise RuntimeError("Invalid date %s" % info)


HANDLERS = []

def register(expression=r'(.*)'):
    compiled = re.compile(expression)
    def chain(func):
        @functools.wraps(func)
        def wrapper(string):
            match = compiled.match(string)
            if match:
                return func(today(), *match.groups())
        HANDLERS.append(wrapper)
        return wrapper
    return chain

def relative_date(value):
    value = value.lower()
    if isinstance(value, str):
        result = next(filter(None, (h(value) for h in HANDLERS)), None)
        if result:
            return result
    return absolute_date(value)

@register(r'today')
def match_today(today):
    return today

@register(r'tomorrow')
def match_tomorrow(today):
    return today + TimeDelta(1)

@register(r'^(\d+)\s+days?$')
def match_days(today, days):
    return today + TimeDelta(int(days))

@register(r'^(\d+)d$')
def match_d(today, days):
    return today + TimeDelta(int(days))

@register(r'^(\d{1,2})$')
def match_day_num(today, day):
    if today.day < int(day):
        return Date(today.year, today.month, int(day))
    else:
        year = today.year + (today.month // 12)
        month = (today.month % 12) + 1
        return Date(year, month, int(day))

@register(r'(\d{1,2})(?:\-|\/)(\d{1,2})$')
def match_month_day_num(today, month, day):
    if today < Date(today.year, int(month), int(day)):
        return Date(today.year, int(month), int(day))
    else:
        return Date(today.year + 1, int(month), int(day))

@register(r'^(mon?|tue?s?|wed?n?e?s?|thu?r?s?|fri?|sat?u?r?|sun?)(?:day)?$')
def match_weekday(today, weekday):
    index = ['mo','tu','we','th','fr','sa','su'].index(weekday[0:2])
    days = (6 - today.weekday() + index) % 7 + 1
    return today + TimeDelta(days)



from unittest import TestCase
from unittest.mock import MagicMock

from busy.commander import Commander
from busy.file_system_root import FilesystemRoot


class TestOneQueueAtATime(TestCase):

    def test_dont_read_plans_when_reading_tasks(self):
        m = FilesystemRoot()
        m.get_queue = MagicMock()
        c = Commander(m)
        c.handle('list')
        try:
            m.get_queue.assert_called_once_with(None)
        except:
            self.fail()

    def test_dont_read_tasks_when_reading_plans(self):
        m = FilesystemRoot()
        m.get_queue = MagicMock()
        c = Commander(m)
        c.handle('list','--queue','plans')
        try:
            m.get_queue.assert_called_once_with('plans')
        except:
            self.fail()

    def test_dont_read_plans_when_pop(self):
        m = FilesystemRoot()
        m.get_queue = MagicMock()
        c = Commander(m)
        c.handle('pop')
        try:
            m.get_queue.assert_called_once_with(None)
        except:
            self.fail()

    def test_save_not_called_if_not_needed(self):    
        m = FilesystemRoot()
        m._save = MagicMock()
        c = Commander(m)
        c.handle('get')
        m._save.assert_not_called()





import sys
import unittest

from pathlib import Path

# sys.path[0] = str(Path(__file__).parent.parent.resolve())
#
# def test(filename):
#     if filename == '__main__':
#         unittest.main()

def foo():
    print("Foo")



from argparse import ArgumentParser

from .file_system_root import FilesystemRoot


class Bash:

    def __init__(self, *input):
        self._parser = ArgumentParser()
        self._parser.add_argument('--root', action='store')
        known, unknown = self._parser.parse_known_args(input)
        self.commands = unknown
        if known.root:
            self.root = FilesystemRoot(known.root)
        else:
            self.root = FilesystemRoot()



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandDefer(TestCase):

    def test_defer(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            c.handle('defer','2','--to','2019-09-06')
            o = p.read_text()
            self.assertEqual(o, 'a\nc\nd\n')
            o2 = Path(t, 'plans.txt').read_text()
            self.assertEqual(o2, '2019-09-06|b\n')

    def test_defer_for(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            c.handle('defer','2','--for','2019-09-06')
            o = p.read_text()
            self.assertEqual(o, 'a\nc\nd\n')
            o2 = Path(t, 'plans.txt').read_text()
            self.assertEqual(o2, '2019-09-06|b\n')

    def test_defer_days(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c = Commander(FilesystemRoot(t))
                c.handle('defer','2','--for','1 day')
                o = p.read_text()
                self.assertEqual(o, 'a\nc\nd\n')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-12|b\n')

    def test_defer_d(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c = Commander(FilesystemRoot(t))
                c.handle('defer','2','--for','5d')
                o = p.read_text()
                self.assertEqual(o, 'a\nc\nd\n')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-16|b\n')

    def test_defer_with_input(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('2019-08-24')):
                c.handle('defer')
                o = p.read_text()
                self.assertEqual(o, 'b\n')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-08-24|a\n')

    def test_default_tomorrow(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                with mock.patch('sys.stdin', StringIO(' ')):
                    c.handle('defer')
                    o2 = Path(t, 'plans.txt').read_text()
                    self.assertEqual(o2, '2019-02-12|a\n')

    def test_tuesday(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,14)):
                c.handle('defer','--to','tue')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-19|a\n')

    def test_thursday(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,14)):
                c.handle('defer','--to','thursday')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-21|a\n')

    def test_capital_days(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,14)):
                c.handle('defer','--to','THURS')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-21|a\n')

    def test_slashes(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a')
            c = Commander(FilesystemRoot(t))
            c.handle('defer','--to','2019/09/06')
            o2 = Path(t, 'plans.txt').read_text()
            self.assertEqual(o2, '2019-09-06|a\n')



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandList(TestCase):

    def test_list(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            c.handle('add','--task','a')
            o = c.handle('list')
            self.assertEqual(o, '     1  a')

    def test_list_plans(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'plans.txt')
            p.write_text('2019-01-04|g\n2019-02-05|p')
            c = Commander(FilesystemRoot(t))
            o = c.handle('list','--queue','plans')
            self.assertEqual(o, '     1  2019-01-04  g\n     2  2019-02-05  p')

    def test_list_with_criteria(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            o = c.handle('list','2','4')
            self.assertEqual(o, '     2  b\n     4  d')




from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest

import busy.dateparser

class TestDate(TestCase):

    @mock.patch('busy.dateparser.today', lambda : Date(2019,2,11))
    def test_today(self):
        t = busy.dateparser.relative_date('today')
        self.assertEqual(t, Date(2019, 2, 11))




from busy.commander import Command
import busy.plugins.todo


class TodoCommand(Command):

    def execute(self, parsed):
        return self.execute_todo(parsed, self._root.get_queue(busy.plugins.todo.TodoQueue.key))

    def execute_todo(self, parsed, queue):  # pragma: no cover
        pass





from unittest import TestCase
from pathlib import Path
from tempfile import TemporaryDirectory

from busy.plugins.todo import TodoQueue
from busy.plugins.todo import PlanQueue
from busy.queue import Queue
from busy.plugins.todo import Task
from busy.file import File
from busy.item import Item

class TestFile(TestCase):

    def test_load_items(self):
        with TemporaryDirectory() as d:
            p = Path(d) / 'x.txt'
            p.write_text('a\nb\n')
            f = File(p)
            q = f.read()
            self.assertIsInstance(q[0], Item)
            self.assertEqual(q[0].description, 'a')

    def test_load_if_not_there(self):
        with TemporaryDirectory() as d:
            f = File(Path(d) / 'y.txt')
            q = f.read()
            self.assertIsInstance(q, list)
            self.assertEqual(len(q), 0)

    def test_save_items(self):
        with TemporaryDirectory() as d:
            p = Path(d) / 'z.txt'
            f1 = File(p)
            i = Item('a')
            f1.save(i)
            f2 = File(p)
            self.assertEqual(str(f2.read()[0]), 'a')

    # def test_plan_file_format(self):
    #     with TemporaryDirectory() as d:
    #         p = Path(d) / 'plans.txt'
    #         p.write_text('2018-12-01|a\n2018-12-09|b')
    #         q = File(p, PlanQueue).queue
    #         self.assertEqual(q.get().date.month, 12)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandDelete(TestCase):

    def test_delete(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            c.handle('delete','--yes','3-')
            o = p.read_text()
            self.assertEqual(o, 'a\nb\n')

    def test_delete_with_input_confirmation_yes(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('Y')):
                c.handle('delete','3-')
                o = p.read_text()
                self.assertEqual(o, 'a\nb\n')

    def test_delete_with_input_confirmation_no(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('no')):
                c.handle('delete','3-')
                o = p.read_text()
                self.assertEqual(o.strip(), 'a\nb\nc\nd')

    def test_delete_outputs_before_confirmation(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\n')
            c = Commander(FilesystemRoot(t))
            o = StringIO()
            with mock.patch('sys.stdin', StringIO('Y')):
                with mock.patch('sys.stdout', o):
                    c.handle('delete', '1')
                    self.assertEqual(o.getvalue(), 'a\nDelete? (Y/n) ')

    def test_delete_defaults_to_first_task_only(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('Y')):
                c.handle('delete')
                o = p.read_text()
                self.assertEqual(o, 'b\n')



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest
from unittest.mock import Mock

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandManage(TestCase):

    def test_manage_launches_editor(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\n')
            with mock.patch('busy.editor', lambda x: 'b\n'):
                c = Commander(FilesystemRoot(t))
                o = c.handle('manage')
                f = p.read_text()
                self.assertEqual(f, 'b\n')

    def test_manage_includes_newline_at_end(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\n')
            m = Mock()
            m.return_value = 'b\n'
            with mock.patch('busy.editor', m):
                c = Commander(FilesystemRoot(t))
                o = c.handle('manage')
                m.assert_called_with('a\n')

    def test_leave_tasks_in_place(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            with mock.patch('busy.editor', lambda x: 'a\n'):
                c = Commander(FilesystemRoot(t))
                o = c.handle('manage', '1')
                f = p.read_text()
                self.assertEqual(f, 'a\nb\n')

    def test_last_record(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\n')
            with mock.patch('busy.editor', lambda x: 'c\n'):
                c = Commander(FilesystemRoot(t))
                o = c.handle('manage', '-')
                f = p.read_text()
                self.assertEqual(f, 'a\nc\n')



from busy.commander import Commander
from . import TodoCommand


class ActivateCommand(TodoCommand):

    command = 'activate'

    @classmethod
    def register(self, parser):
        parser.add_argument('--today', '-t', action='store_true')

    def execute_todo(self, parsed, queue):
        if hasattr(parsed, 'today') and parsed.today:
            queue.activate(today=True)
        else:
            queue.activate(*parsed.criteria)


Commander.register(ActivateCommand)





from ..commander import QueueCommand
from ..commander import Commander


class ListCommand(QueueCommand):

    command = 'list'

    def execute_on_queue(self, parsed, queue):
        itemlist = queue.list(*parsed.criteria)
        return self._list(queue, itemlist)


Commander.register(ListCommand)



from unittest import TestCase
import datetime

from busy.plugins.todo import Task
from busy.plugins.todo import Plan

class TestTask(TestCase):

    def test_task(self):
        t = Task('b')
        self.assertIsInstance(t, Task)

    def test_description_as_string(self):
        t = Task('a')
        self.assertEqual(t.description, 'a')

    def test_task_requires_description(self):
        with self.assertRaises(AssertionError):
            t = Task('')

    def test_plan(self):
        t = Task('a')
        d = datetime.date(2018,12,1)
        t.as_plan(d)

    def test_plan_as_tuple(self):
        t = Task('b')
        p = t.as_plan((2018,12,1))
        self.assertEqual(p.date.month, 12)

    def test_plan_requires_date(self):
        t = Task('c')
        with self.assertRaises(RuntimeError):
            t.as_plan(54)

    def test_plan_with_date_as_string(self):
        t = Task('d')
        p = t.as_plan('2019-04-05')
        self.assertEqual(p.date.day, 5)

    def test_create_plan_with_dict(self):
        t = Plan(date=(2019, 4,15), description='a')
        self.assertEqual(t.date.month, 4)
 
    def test_tags(self):
        t = Task('f #a')
        self.assertEqual(t.tags, ['a'])

    def test_project(self):
        t = Task("k #oNe #two")
        self.assertEqual(t.project, 'one')

    def test_resource(self):
        t = Task("a at https://b.us #f")
        self.assertEqual(t.resource, "https://b.us")
    
    def test_resource_without_at(self):
        t = Task("e")
        self.assertEqual(t.resource, "")

    def test_without_resource(self):
        t = Task("a at https://b.us #f")
        self.assertEqual(t.without_resource, 'a #f')

    def test_without_resource_without_at(self):
        t = Task("f")
        self.assertEqual(t.without_resource, "f")



from ..commander import QueueCommand
from ..commander import Commander


class GetCommand(QueueCommand):

    command = 'get'

    def execute_on_queue(self, parsed, queue):
        if parsed.criteria:
            message = ("The `get` command only returns the top item - "
                       "repeat without criteria")
            raise RuntimeError(message)
        else:
            return str(queue.get() or '')


Commander.register(GetCommand)



from busy.commander import Commander
from . import TodoCommand


class FinishCommand(TodoCommand):

    command = 'finish'

    @classmethod
    def register(self, parser):
        parser.add_argument('--yes', action='store_true')

    def execute_todo(self, parsed, queue):
        tasklist = queue.list(*parsed.criteria or [1])
        indices = [i[0]-1 for i in tasklist]
        if self.is_confirmed(parsed, tasklist, 'Finish', 'Finishing'):
            queue.finish(*indices)


Commander.register(FinishCommand)





from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestAlternateQueue(TestCase):

    def test_get_from(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'a.txt')
            p.write_text('b\n')
            c = Commander(FilesystemRoot(t))
            o = c.handle('get','--queue','a')
            self.assertEqual(o, 'b')

    def test_default_queue(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('b\n')
            c = Commander(FilesystemRoot(t))
            o = c.handle('get')
            self.assertEqual(o, 'b')

    def test_add(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('g')):
                c.handle('add','--queue','j')
                x = Path(t, 'j.txt').read_text()
                self.assertEqual(x, 'g\n')

    def test_drop(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'u.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            c.handle('drop','2','4','--queue','u')
            o = p.read_text()
            self.assertEqual(o, 'a\nc\nb\nd\n')

    def test_delete(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'w.txt')
            p.write_text('a\nb\nc\nd')
            c = Commander(FilesystemRoot(t))
            c.handle('delete','--yes','3-', '--queue', 'w')
            o = p.read_text()
            self.assertEqual(o, 'a\nb\n')

    def test_manage(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'y.txt')
            p.write_text('a\n')
            with mock.patch('busy.editor', lambda x: 'b\n'):
                c = Commander(FilesystemRoot(t))
                c.handle('manage', '--queue', 'y')
                f = p.read_text()
                self.assertEqual(f, 'b\n')



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from io import StringIO
from datetime import date as Date
import unittest
import random

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot


class TestCommandShuffle(TestCase):

    def test_shuffle(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb\nc\n')
            random.seed(1)
            c = Commander(FilesystemRoot(t))
            c.handle('shuffle')
            f = p.read_text()
            self.assertEqual(f, 'b\nc\na\n')




# General-purpose queue of items
# Method calls on queue use indices that start at 1

import random

from .selector import Selector
from .item import Item


class Queue:

    itemclass = Item
    default_key = 'default'

    def __init__(self, manager=None, items=[]):
        self.manager = manager
        self.changed = False
        self._items = [self.itemclass.create(i) for i in items if i]

    def all(self):
        return self._items

    def count(self):
        return len(self._items)

    # Add new items. Always makes them the right class. Also does inserts.

    def add(self, *items, index=None):
        newitems = [self.itemclass.create(i) for i in items if i]
        index = len(self._items) if index is None else index
        self._items[index:index] = newitems
        self.changed = True

    # Replace existing items at the indices provided. Also inserts if the
    # indices run out. Does not create items. Would be good to combine this
    # with the add method.

    def replace(self, indices, newvalues):
        while newvalues and indices:
            self._items[indices.pop(0)] = newvalues.pop(0)
        while indices:
            del self._items[indices.pop()]
        self._items.extend(newvalues)
        self.changed = True

    def get(self, index=1):
        return self._items[index-1] if self._items else None

    def select(self, *criteria):
        selector = Selector(criteria)
        return selector.indices(self._items)

    def _split(self, *criteria):
        return self._split_by_indices(*self.select(*criteria))

    def _split_by_indices(self, *indices):
        inlist = [t for i, t in enumerate(self._items) if i in indices]
        outlist = [t for i, t in enumerate(self._items) if i not in indices]
        return (inlist, outlist)

    def pop(self, *criteria):
        hilist, lolist = self._split(*criteria or [len(self._items)])
        self._items = hilist + lolist
        self.changed = True

    def drop(self, *criteria):
        lolist, hilist = self._split(*criteria or [1])
        self._items = hilist + lolist
        self.changed = True

    def delete(self, *criteria):
        self._items = self._split(*criteria)[1]
        self.changed = True

    def delete_by_indices(self, *indices):
        self._items = self._split_by_indices(*indices)[1]
        self.changed = True

    def list(self, *criteria):
        return [(i+1, self._items[i]) for i in self.select(*criteria)]

    @property
    def strings(self):
        return [str(i) for i in self._items]

    def manage(self, *criteria, editor):
        itemlist = self.list(*criteria)
        indices = [i[0]-1 for i in itemlist]
        items = [i[1] for i in itemlist]
        new_items = editor(self.itemclass, *items)
        self.replace(indices, new_items)

    def shuffle(self, *criteria):
        itemlist = self.list(*criteria)
        indices = [i[0]-1 for i in itemlist]
        items = [i[1] for i in itemlist]

        random.shuffle(items)

        self.replace(indices, items)

    @classmethod
    def register(self, queueclass, default=False):
        if not hasattr(self, '_classes'):
            self._classes = {}
        self._classes[queueclass.key] = queueclass
        if default:
            self.default_key = queueclass.key

    @classmethod
    def subclass(self, key):
        return self._classes.get(key) or self

    # Get all the tags for the queue, as a set

    def tags(self):
        master = [i.tags for i in self._items]
        return {item for sublist in master for item in sublist}


from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.commander import Commander
from busy.file_system_root import FilesystemRoot


class TestCommandTags(TestCase):

    def test_tags(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a #b\nc #d\ne #b\n')
            c = Commander(FilesystemRoot(t))
            o = c.handle('tags')
            self.assertEqual(o, 'b\nd')



from csv import DictReader
from csv import DictWriter
from pathlib import Path

from .item import Item


def read_items(fileish, itemclass=Item):
    reader = DictReader(fileish, itemclass.schema, delimiter="|")
    return [itemclass.create(i) for i in reader if i]


def write_items(fileish, *items):
    schema = items[0].schema
    writer = DictWriter(fileish, schema, delimiter="|")
    for item in items:
        values = dict([(f, getattr(item, f)) for f in schema])
        writer.writerow(values)


class File:

    def __init__(self, path):
        self._path = path

    def read(self, itemclass=Item):
        if self._path.is_file():
            with open(self._path) as datafile:
                return read_items(datafile, itemclass)
        return []

    def save(self, *items):
        if items:
            with open(self._path, 'w') as datafile:
                write_items(datafile, *items)
        else:
            Path(self._path).write_text('')



from busy import dateparser

from busy.commander import Commander
from . import TodoCommand


class DeferCommand(TodoCommand):

    command = 'defer'

    @classmethod
    def register(self, parser):
        parser.add_argument('--to', '--for', dest='time_info')

    def execute_todo(self, parsed, queue):
        tasklist = queue.list(*parsed.criteria or [1])
        # indices = [i[0]-1 for i in tasklist]
        if hasattr(parsed, 'time_info') and parsed.time_info:
            time_info = parsed.time_info
        else:
            print('\n'.join([str(i[1]) for i in tasklist]))
            time_info = input('Defer to [tomorrow]: ').strip() or 'tomorrow'
        queue.defer(dateparser.relative_date(time_info), *parsed.criteria)


Commander.register(DeferCommand)





from unittest import TestCase
from unittest import mock
from tempfile import TemporaryDirectory
import os

from busy.bash import Bash
from busy.file_system_root import FilesystemRoot


class TestBash(TestCase):

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        os.environ['BUSY_ROOT'] = self.tempdir.name

    def tearDown(self):
        self.tempdir.cleanup()

    def test_commands(self):
        b = Bash('a', 'b')
        self.assertEqual(b.commands, ['a', 'b'])

    def test_default_storage(self):
        b = Bash('a', 'b')
        self.assertIsInstance(b.root, FilesystemRoot)

    def test_root_option(self):
        with TemporaryDirectory() as d:
            b = Bash('--root', d)
            self.assertEqual(str(b.root._path), d)



from busy.commander import Commander
from . import TodoCommand
from ....commands.manage_command import text_editor


class StartCommand(TodoCommand):

    command = 'start'

    @classmethod
    def register(self, parser):
        parser.add_argument('project', action='store', nargs='?')

    def execute_todo(self, parsed, queue):
        if parsed.criteria:
            raise RuntimeError('Start takes only an optional project name')
        queue.activate(today=True)
        if queue.count() < 1:
            raise RuntimeError('There are no active tasks')
        project = parsed.project or queue.get().project
        if not project:
            raise RuntimeError('The `start` command required a project')
        queue.manage(project, editor=text_editor)
        queue.pop(project)


Commander.register(StartCommand)



# A generic item from a queue, with a description and possibly tags


class Item:

    schema = ['description']
    listfmt = "{1.description}"

    def __init__(self, description=None):
        assert isinstance(description, str)
        assert description
        self._description = description

    def __str__(self):
        return self._description

    @property
    def description(self):
        return self._description

    @classmethod
    def create(self, value):
        if isinstance(value, self):
            return value
        elif isinstance(value, str):
            return self(value)
        elif isinstance(value, dict):
            return self(**value)

    @property
    def tags(self):
        words = self.description.split()
        return [w[1:].lower() for w in words if w.startswith('#')]





class Selector:

    CRITERIUM_TYPES = []

    @classmethod
    def add_criterium_type(self, criterium_type):
        self.CRITERIUM_TYPES.append(criterium_type)

    @classmethod
    def get_criterium(self, word):
        for criterium_type in self.CRITERIUM_TYPES:
            if criterium_type.match(word):
                return criterium_type(word)

    def __init__(self, args):
        self.criteria = [c for c in [self.get_criterium(w) for w in args] if c]

    def hit(self, index, value, count):
        if self.criteria:
            return any([c.hit(index, value, count) for c in self.criteria])
        else:
            return True

    def indices(self, elements):
        enumeration = enumerate(elements)
        count = len(elements)
        return [i for i, t in enumeration if self.hit(i, t, count)]


class NumberCriterium:

    def match(word):
        return str(word).isdigit()

    def __init__(self, word):
        self.index = int(word) - 1

    def hit(self, index, value, count):
        return index == self.index


Selector.add_criterium_type(NumberCriterium)


class RangeCriterium:

    def match(word):
        return '-' in str(word)

    def __init__(self, word):
        split = word.split('-')
        self.start = int(split[0]) - 1 if split[0] else None
        self.end = int(split[1]) - 1 if split[1] else None

    def hit(self, index, value, count):
        if self.start is None and self.end is None:
            return index == count - 1
        elif self.end is None:
            return index >= self.start
        else:
            return index >= self.start and index <= self.end


Selector.add_criterium_type(RangeCriterium)


class TagCriterium:

    def match(word):
        return str(word).isidentifier()

    def __init__(self, word):
        self.tag = str(word).lower()

    def hit(self, index, value, count):
        return self.tag in value.tags


Selector.add_criterium_type(TagCriterium)


class FunctionCriterium:

    def match(arg):
        return callable(arg)

    def __init__(self, arg):
        self.func = arg

    def hit(self, index, value, count):
        return self.func(value)


Selector.add_criterium_type(FunctionCriterium)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest

from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandFinish(TestCase):

    def test_finish(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = Path(t, 'done.txt').read_text()
                self.assertEqual(o, '2019-02-11|a\n')

    def test_confirmation(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\n')
            c = Commander(FilesystemRoot(t))
            o = StringIO()
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                with mock.patch('sys.stdin', StringIO('Y')):
                    with mock.patch('sys.stdout', o):
                        c.handle('finish')
                        o = Path(t, 'done.txt').read_text()
                        self.assertEqual(o, '2019-02-11|a\n')

    def test_followon(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>b\n')
            c = Commander(FilesystemRoot(t))
            c.handle('finish','--yes')
            o = p.read_text()
            self.assertEqual(o, 'b\n')

    def test_followon_records_only_first_task_as_done(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>b\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = Path(t, 'done.txt').read_text()
                self.assertEqual(o, '2019-02-11|a\n')

    def test_followon_with_alt_chars(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a  --> b\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, 'b\n')
                o2 = Path(t, 'done.txt').read_text()
                self.assertEqual(o2, '2019-02-11|a\n')

    def test_repeat(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat in 2 days\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-13|a>repeat in 2 days\n')
                o3 = Path(t, 'done.txt').read_text()
                self.assertEqual(o3, '2019-02-11|a\n')

    def test_repeat_parse_fail(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat in x\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                with self.assertRaises(RuntimeError):
                    c.handle('finish','--yes')

    def test_repeat_tomorrow(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat tomorrow\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-12|a>repeat tomorrow\n')
                o3 = Path(t, 'done.txt').read_text()
                self.assertEqual(o3, '2019-02-11|a\n')

    def test_repeat_day_template(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat on 16\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-02-16|a>repeat on 16\n')

    def test_repeat_day_template_next_month(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat on 4\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-03-04|a>repeat on 4\n')

    def test_repeat_day_template_next_year(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat on 4\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,12,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2020-01-04|a>repeat on 4\n')

    def test_repeat_month_day_template(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat on 7-16\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2019-07-16|a>repeat on 7-16\n')

    def test_repeat_month_day_next_year(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>repeat on 7-16\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,9,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2020-07-16|a>repeat on 7-16\n')

    def test_finish_caps(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a>Repeat ON 7-16\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,9,11)):
                c.handle('finish','--yes')
                o = p.read_text()
                self.assertEqual(o, '')
                o2 = Path(t, 'plans.txt').read_text()
                self.assertEqual(o2, '2020-07-16|a>Repeat ON 7-16\n')



from ..commander import QueueCommand
from ..commander import Commander
from ..ui.edit_task_ui import EditTaskUi


class AddCommand(QueueCommand):

    command = 'add'

    @classmethod
    def register(self, parser):
        super().register(parser)
        parser.add_argument('item', nargs='?')
        parser.add_argument('--interactive', action='store_true')

    def execute_on_queue(self, parsed, queue):
        if parsed.interactive:  # pragma: no cover
            frame = EditTaskUi(lambda task: queue.add(task))
        else:
            if hasattr(parsed, 'item') and parsed.item:
                item = parsed.item
            else:
                item = input('Item: ')
            queue.add(item)


Commander.register(AddCommand)



from ..commander import QueueCommand
from ..commander import Commander


class TagsCommand(QueueCommand):

    command = 'tags'

    def execute_on_queue(self, parsed, queue):
        return '\n'.join(sorted(queue.tags()))


Commander.register(TagsCommand)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

@mock.patch('busy.editor', lambda x: x)
class TestCommandStart(TestCase):

    def test_start_works_for_named_project(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb #g\nc #k\nd #g')
            c = Commander(FilesystemRoot(t))
            o = c.handle('start','g')
            f = p.read_text()
            self.assertEqual(f, 'b #g\nd #g\na\nc #k\n')

    def test_start_fails_if_criteria(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb #g\nc #k\nd #g')
            c = Commander(FilesystemRoot(t))
            with self.assertRaises(RuntimeError):
                c.handle('start','g','4')

    def test_start_pops_tasks_for_current_project(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a #k\nb #g\nc #k\nd #g')
            c = Commander(FilesystemRoot(t))
            c.handle('start')
            o = p.read_text()
            self.assertEqual(o, 'a #k\nc #k\nb #g\nd #g\n')

    def test_start_activates_today(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'plans.txt')
            p.write_text('2018-09-04|x #g\n2019-03-25|y #g\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019, 2, 11)):
                o = c.handle('start','g')
                self.assertEqual(p.read_text(), '2019-03-25|y #g\n')
                p2 = Path(t, 'tasks.txt')
                self.assertEqual(p2.read_text(), 'x #g\n')

    def test_start_fails_gracefully_if_no_todos(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            with self.assertRaises(RuntimeError):
                c.handle('start', 'g')

    def test_start_fails_gracefully_if_no_project_at_top(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'tasks.txt')
            p.write_text('a\nb #g')
            c = Commander(FilesystemRoot(t))
            with self.assertRaises(RuntimeError):
                c.handle('start')



from ..commander import QueueCommand
from ..commander import Commander


class DeleteCommand(QueueCommand):

    command = 'delete'

    @classmethod
    def register(self, parser):
        super().register(parser)
        parser.add_argument('--yes', action='store_true')

    def execute_on_queue(self, parsed, queue):
        itemlist = queue.list(*parsed.criteria or [1])
        indices = [i[0]-1 for i in itemlist]
        if self.is_confirmed(parsed, itemlist, 'Delete', 'Deletion'):
            queue.delete_by_indices(*indices)


Commander.register(DeleteCommand)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date
import unittest

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandActivate(TestCase):

    def test_activate(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'plans.txt')
            p.write_text('2018-09-04|a\n')
            c = Commander(FilesystemRoot(t))
            c.handle('activate','1')
            self.assertEqual(p.read_text(), '')
            p2 = Path(t, 'tasks.txt')
            self.assertEqual(p2.read_text(), 'a\n')

    def test_activate_with_today_option(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'plans.txt')
            p.write_text('2018-09-04|a\n2019-03-25|b\n')
            c = Commander(FilesystemRoot(t))
            with mock.patch('busy.dateparser.today', lambda : Date(2019,2,11)):
                c.handle('activate','--today')
                self.assertEqual(p.read_text(), '2019-03-25|b\n')
                p2 = Path(t, 'tasks.txt')
                self.assertEqual(p2.read_text(), 'a\n')

    def test_pop(self):
        with TemporaryDirectory() as t:
            p1 = Path(t, 'tasks.txt')
            p1.write_text('x\n')
            p2 = Path(t, 'plans.txt')
            p2.write_text('2018-09-04|a\n')
            c = Commander(FilesystemRoot(t))
            c.handle('activate','1')
            self.assertEqual(p1.read_text(), 'a\nx\n')

    def test_activate_no_args_does_nothing(self):
        with TemporaryDirectory() as t:
            p = Path(t, 'plans.txt')
            p.write_text('2018-09-04|a\n')
            c = Commander(FilesystemRoot(t))
            c.handle('activate')
            self.assertEqual(p.read_text(), '2018-09-04|a\n')
            p2 = Path(t, 'tasks.txt')
            self.assertFalse(p2.exists())



from ..commander import Command
from ..commander import Commander


class QueuesCommand(Command):

    command = 'queues'

    def execute(self, parsed):
        return '\n'.join(sorted(self._root.queue_names))


Commander.register(QueuesCommand)



from ..commander import QueueCommand
from ..commander import Commander


class DropCommand(QueueCommand):

    command = 'drop'


Commander.register(DropCommand)


class PopCommand(QueueCommand):

    command = 'pop'


Commander.register(PopCommand)



import sys

from .file_system_root import FilesystemRoot
from .commander import Commander
from .bash import Bash


def main():
    bash = Bash(*sys.argv[1:])
    try:
        output = Commander(bash.root).handle(*bash.commands)
    except (RuntimeError, ValueError) as error:
        output = f"Error: {str(error)}"
    if output:
        print(output)


if __name__ == '__main__':
    main()



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot


class TestCommander(TestCase):

    def test_with_root_param(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            o = c.handle('add', '--task', 'a')
            x = Path(t, 'tasks.txt').read_text()
            self.assertEqual(x, 'a\n')

    def test_version_info(self):
        with TemporaryDirectory() as t:
            with mock.patch('sys.version_info', (1, 2, 3)):
                with self.assertRaises(RuntimeError):
                    Commander(FilesystemRoot(t))



from unittest import TestCase
import busy
from unittest import mock
from pathlib import Path

class MockSubprocess:

    @staticmethod
    def run(arg):
        Path(arg[-1]).write_text('v')

class MockShUtil:

    @staticmethod
    def which(arg):
        return False


class TestEditor(TestCase):

    def test_editor(self):
        with mock.patch('busy.subprocess', MockSubprocess):
            x = busy.editor('a')
            self.assertEqual(x, 'v')

    def test_no_command(self):
        with mock.patch('busy.shutil', MockShUtil):
            with self.assertRaises(RuntimeError):
                busy.editor('a')



from busy.commander import Commander
from . import TodoCommand


class ResourceCommand(TodoCommand):

    command = "resource"

    def execute_todo(self, parsed, queue):
        if parsed.criteria:  # pragma: nocover
            message = ("The `resource` command only returns the top item - "
                       "repeat without criteria")
            raise RuntimeError(message)
        else:
            return str(queue.resource() or '')


Commander.register(ResourceCommand)





from io import StringIO

from ..commander import QueueCommand
from ..commander import Commander
from ..file import write_items
from ..file import read_items
from ..ui.edit_task_ui import EditTaskUi
import busy


def text_editor(itemclass, *items):
    sio = StringIO(newline=None)
    write_items(sio, *items)
    before = sio.getvalue()
    sio.close()
    edited = busy.editor(before)
    sio2 = StringIO(edited)
    new_items = read_items(sio2, itemclass)
    sio2.close()
    return new_items


def single_task_ui_editor(itemclass, *items):  # pragma: no cover
    assert len(items) == 1
    ui = EditTaskUi(initial_value=str(items[0]))
    return [itemclass(ui.value)]


class ManageCommand(QueueCommand):

    command = 'manage'

    @classmethod
    def register(self, parser):
        super().register(parser)
        parser.add_argument('--interactive', action='store_true')

    def execute_on_queue(self, parsed, queue):
        if parsed.interactive:  # pragma: no cover
            editor = single_task_ui_editor
        else:
            editor = text_editor
        queue.manage(*parsed.criteria, editor=editor)


Commander.register(ManageCommand)



from unittest import TestCase
import unittest
import random

from busy.queue import Queue


class TestQueue(TestCase):

    def test_list(self):
        q = Queue()
        q.add('a')
        q.add('b')
        i = q.list()
        self.assertEqual(i[0][0], 1)
        self.assertEqual(str(i[0][1]), 'a')

    def test_list_with_criteria(self):
        q = Queue()
        q.add('a','b','c')
        i = q.list(1,3)
        self.assertEqual(i[1][0], 3)
        self.assertEqual(str(i[1][1]), 'c')

    def test_select_multiple(self):
        q = Queue()
        q.add('a','b','c')
        t = q.select(1,3)
        self.assertEqual(len(t), 2)

    def test_pop_multiple(self):
        q = Queue()
        q.add('a','b','c','d')
        q.pop(2,4)
        self.assertEqual(str(q.get(2)), 'd')

    def test_pop_if_nothing(self):
        q = Queue()
        q.pop()
        self.assertEqual(q.count(), 0)

    def test_drop(self):
        q = Queue()
        q.add('a','b','c','d')
        q.drop()
        self.assertEqual(str(q.get(1)), 'b')

    def test_delete(self):
        q = Queue()
        q.add('a','b','c')
        q.delete(2,3)
        self.assertEqual(q.count(), 1)

    def test_create_with_array_of_dict(self):
        q = Queue()
        q.add({'description':'a'})
        self.assertEqual(str(q.get()), 'a')

    def test_list_with_range(self):
        q = Queue()
        q.add('a','b','c')
        t = q.select('2-3')
        self.assertEqual(len(t), 2)

    def test_list_with_range_ending(self):
        q = Queue()
        q.add('a','b','c','d')
        t = q.list('2-')
        self.assertEqual(len(t), 3)
        self.assertEqual(str(t[0][1]), 'b')

    def test_list_only_end(self):
        q = Queue()
        q.add('a', 'b', 'c', 'd')
        t = q.list('-')
        self.assertEqual(len(t), 1)
        self.assertEqual(str(t[0][1]), 'd')

    def test_partial_tags_dont_match(self):
        q = Queue()
        q.add('a #b', 'b #busy')
        t = q.list('b')
        self.assertEqual(len(t), 1)

    def test_tags(self):
        q = Queue()
        q.add('a #b', 'c #d', 'e #b')
        t = q.tags()
        self.assertEqual(t, {'b', 'd'})

    def test_shuffle(self):
        q = Queue()
        q.add('a', 'b', 'c')
        random.seed(1)
        q.shuffle()
        s = q.list()
        self.assertEqual(str(s[0][1]), 'b')
        self.assertEqual(str(s[1][1]), 'c')



from unittest import TestCase

from busy.queue import Queue

class TestQueueReplace(TestCase):

    def test_replace_one(self):
        a = Queue()
        a.add('a','b','c')
        a.replace([1], ['d'])
        self.assertEqual(str(a.get(2)), 'd')

    def test_replace_multi(self):
        a = Queue()
        a.add('a','b','c','d')
        a.replace([1,2], ['1','2'])
        self.assertEqual(a.strings, ['a','1','2','d'])

    def test_replace_extra_index(self):
        a = Queue()
        a.add('a','b','c','d')
        a.replace([1,2], ['1'])
        self.assertEqual(a.strings, ['a','1','d'])



import re

from ...queue import Queue
from ...item import Item
from ...commander import Command
from ...commander import Commander
from busy import dateparser
from .commands import TodoCommand

MARKER = re.compile(r'\s*\-*\>\s*')
REPEAT = re.compile(r'^\s*repeat(?:\s+[io]n)?\s+(.+)\s*$', re.I)
RESOURCE = re.compile(r'\s+at\s+(\S+)')


class Task(Item):

    def __init__(self, description=None):
        super().__init__(description)

    def as_plan(self, date):
        return Plan(self.description, date)

    def as_done(self, date):
        return DoneTask(self._marker_split[0], date)

    def as_followon(self):
        if len(self._marker_split) > 1:
            description = self._marker_split[1]
            if not REPEAT.match(description):
                return Task(description)

    def as_repeat(self):
        if len(self._marker_split) > 1:
            match = REPEAT.match(self._marker_split[1])
            if match:
                date = dateparser.relative_date(match.group(1))
                return Plan(self.description, date)

    @property
    def project(self):
        tags = self.tags
        return tags[0] if tags else None

    @property
    def _marker_split(self):
        return MARKER.split(self.description, maxsplit=1)

    @property
    def resource(self):
        match = RESOURCE.search(self.description)
        if match:
            return match.group(1)
        else:
            return ""

    @property
    def without_resource(self):
        split = RESOURCE.split(self.description, maxsplit=1)
        if len(split) > 1:
            return split[0] + split[2]
        else:
            return self.description


class Plan(Item):

    schema = ['date', 'description']
    listfmt = "{1.date:%Y-%m-%d}  {1.description}"

    def __init__(self, description=None, date=None):
        super().__init__(description)
        self._date = dateparser.absolute_date(date)

    @property
    def date(self):
        return self._date

    def as_todo(self):
        return Task(self.description)


class DoneTask(Item):

    schema = ['date', 'description']
    listfmt = "{1.date:%Y-%m-%d}  {1.description}"

    def __init__(self, description=None, date=None):
        super().__init__(description)
        self._date = dateparser.absolute_date(date)

    @property
    def date(self):
        return self._date


def is_today_or_earlier(plan):
    return plan.date <= dateparser.today()


class TodoQueue(Queue):
    itemclass = Task
    key = 'tasks'

    def __init__(self, manager=None, items=[]):
        super().__init__(manager, items)
        self._plans = None
        self._done = None

    @property
    def plans(self):
        if not self._plans:
            if self.manager:
                self._plans = self.manager.get_queue(PlanQueue.key)
            else:
                self._plans = PlanQueue()
        return self._plans

    @property
    def done(self):
        if not self._done:
            if self.manager:
                self._done = self.manager.get_queue('done')
            else:
                self._done = DoneQueue()
        return self._done

    def defer(self, date, *criteria):
        indices = self.select(*(criteria or [1]))
        plans = [self.get(i+1).as_plan(date) for i in indices]
        self.plans.add(*plans)
        self.delete_by_indices(*indices)

    def activate(self, *criteria, today=False):
        if today:
            indices = self.plans.select(is_today_or_earlier)
        elif criteria:
            indices = self.plans.select(*criteria)
        else:
            return
        tasks = [self.plans.get(i+1).as_todo() for i in indices]
        self.add(*tasks, index=0)
        self.plans.delete_by_indices(*indices)

    def finish(self, *indices, date=None):
        if not date:
            date = dateparser.today()
        donelist, keeplist = self._split_by_indices(*indices)
        self._items = keeplist
        self.done.add(*[t.as_done(date) for t in donelist])
        self.add(*filter(None, [t.as_followon() for t in donelist]))
        self.plans.add(*filter(None, [t.as_repeat() for t in donelist]))

    def resource(self, index=1):
        return self._items[index-1].resource if self._items else None

    def get_without_resource(self, index=1):
        return self._items[index-1].without_resource if self._items else None


Queue.register(TodoQueue, default=True)


class PlanQueue(Queue):
    itemclass = Plan
    key = 'plans'


Queue.register(PlanQueue)


class DoneQueue(Queue):
    itemclass = DoneTask
    key = 'done'


Queue.register(DoneQueue)



from busy.commander import Commander
from . import TodoCommand


class GetWithoutResourceCommand(TodoCommand):

    command = "get-without-resource"

    def execute_todo(self, parsed, queue):
        if parsed.criteria:  # pragma: nocover
            message = ("The `get-without-resource` command only returns the"
                       " top item - repeat without criteria")
            raise RuntimeError(message)
        else:
            return str(queue.get_without_resource() or '')


Commander.register(GetWithoutResourceCommand)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.plugins.todo import Task
from busy.commander import Commander
from busy.file_system_root import FilesystemRoot

class TestCommandAdd(TestCase):

    def test_by_parameter(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            c.handle('add','u p')
            x = Path(t, 'tasks.txt').read_text()
            self.assertEqual(x, 'u p\n')

    def test_add_by_input(self):
        with TemporaryDirectory() as t:
            c = Commander(FilesystemRoot(t))
            with mock.patch('sys.stdin', StringIO('g')):
                c.handle('add')
                x = Path(t, 'tasks.txt').read_text()
                self.assertEqual(x, 'g\n')



import subprocess
from tempfile import NamedTemporaryFile
from pathlib import Path
import shutil
import os

PYTHON_VERSION = (3, 6, 5)

COMMANDS = [['sensible-editor'], ['open', '-W']]


def editor(arg):
    with NamedTemporaryFile() as tempfile:
        Path(tempfile.name).write_text(arg)
        command = [os.environ.get('EDITOR')]
        if not command[0] or not shutil.which(command[0]):
            iterator = (c for c in COMMANDS if shutil.which(c[0]))
            command = next(filter(None, iterator), None)
            if not command:
                raise RuntimeError("The manage command required an editor")
        subprocess.run(command + [tempfile.name])
        return Path(tempfile.name).read_text()



from argparse import ArgumentParser
from tempfile import TemporaryDirectory
import sys
from tempfile import TemporaryFile
from pathlib import Path
import importlib
import re

from . import PYTHON_VERSION
import busy


MESSAGE = "Busy requires Python version %i.%i.%i or higher"


class Commander:

    def __init__(self, root):
        if sys.version_info < PYTHON_VERSION:
            raise RuntimeError(MESSAGE % PYTHON_VERSION)
        self._root = root

    def handle(self, *args):
        parsed, remaining = self._parser.parse_known_args(args)
        parsed.criteria = remaining
        if hasattr(parsed, 'command'):
            command = parsed.command(self.root)
            result = command.execute(parsed)
            command.save()
            return result

    @property
    def root(self):
        return self._root

    @classmethod
    def register(self, command_class):
        if not hasattr(self, '_parser'):
            self._parser = ArgumentParser()
            self._subparsers = self._parser.add_subparsers()
        subparser = self._subparsers.add_parser(command_class.command)
        subparser.set_defaults(command=command_class)
        command_class.register(subparser)


class Command:

    def __init__(self, root):
        self._root = root

    def _list(self, queue, tasklist):
        fmtstring = "{0:>6}  " + queue.itemclass.listfmt
        texts = [fmtstring.format(i, t) for i, t in tasklist]
        return '\n'.join(texts)

    def save(self):
        self._root.save()

    def is_confirmed(self, parsed, itemlist, verb, noun):
        if hasattr(parsed, 'yes') and parsed.yes:
            confirmed = True
        else:
            print('\n'.join([str(i[1]) for i in itemlist]))
            confirmed = input(f'{verb}? (Y/n) ').startswith('Y')
        if not confirmed:
            print(f"{noun} must be confirmed")
        return confirmed

    @classmethod
    def register(self, parser):
        pass


class QueueCommand(Command):

    @classmethod
    def register(self, parser):
        parser.add_argument('--queue', nargs=1, dest="queue")

    def execute(self, parsed):
        key = parsed.queue[0] if getattr(parsed, 'queue') else None
        queue = self._root.get_queue(key)
        return self.execute_on_queue(parsed, queue)

    def execute_on_queue(self, parsed, queue):
        method = getattr(queue, self.command)
        result = method(*parsed.criteria)
        return result


def do_import(folder_path, module_base):
    for import_file in folder_path.iterdir():
        if re.match(r'^[^_].*\.py$', import_file.name):
            importlib.import_module(f'{module_base}.{import_file.stem}')


do_import(Path(__file__).parent / 'commands', 'busy.commands')
do_import(Path(__file__).parent / 'plugins' / 'todo', 'busy.plugins.todo')
do_import(Path(__file__).parent / 'plugins' / 'todo' / 'commands', 'busy.plugins.todo.commands')
